// Author: Alexander S. Gerasimov

#include <iostream>


class Counter {
private:
    int c;
public:
    Counter();
    Counter(int i);
    virtual ~Counter();
    virtual void incr();
    void decr();
    int getValue();
    virtual void print();
};


Counter::Counter() {
    std::cout<<"Counter()\n";
    c = 0;
}

Counter::Counter(int i) {
    std::cout<<"Counter("<<i<<")\n";
    if (i < 0) i = 0;
    c = i;
}

Counter::~Counter() { std::cout<<"~Counter()\n"; }

void Counter::incr() { c++; }

void Counter::decr() { if (c > 0) c--; }

int Counter::getValue() { return c; }

void Counter::print() { std::cout<<c<<'\n'; }


class InstrumentedCounter : public Counter {
private:
    int numOfIncrs; // number of increments
public:
    InstrumentedCounter();
    ~InstrumentedCounter();
    virtual void incr();
    int getNumOfIncrs();
    virtual void print();
};


InstrumentedCounter::InstrumentedCounter() : Counter(0) {
    std::cout<<"InstrumentedCounter()\n";
    numOfIncrs = 0;
}

InstrumentedCounter::~InstrumentedCounter() {
    std::cout<<"~InstrumentedCounter()\n";
}

void InstrumentedCounter::incr() {
    Counter::incr();
    numOfIncrs++;
}

int InstrumentedCounter::getNumOfIncrs() { return numOfIncrs; }

void InstrumentedCounter::print() {
    std::cout<<getValue()<<", numOfIncrs="<<numOfIncrs<<'\n';
}


void usePointer(Counter* pcntr) {
    pcntr->incr();
    pcntr->decr();
    pcntr->incr();
    pcntr->print();
}

void useRef(Counter& cntr) {
    cntr.incr();
    cntr.decr();
    cntr.incr();
    cntr.print();
}

int main() {
    {
        InstrumentedCounter cntr;
        usePointer(&cntr);
        useRef(cntr);
    }
    std::cout<<'\n';
    Counter* pcntr = new InstrumentedCounter;
    usePointer(pcntr);
    delete pcntr;
    return 0;
}
